import os, sys, json
from collections import defaultdict

from tipp3.configs import Configs
from tipp3 import get_logger
from tipp3.jobs import TIPPJsonMergerJob

#from tqdm import tqdm
#from tipp3.helpers.general_tools import tqdm_styles

import concurrent.futures

_LOG = get_logger(__name__)

# taxonomic ranks
ranks = ['species', 'genus', 'family', 'order', 'class', 'phylum', 'superkingdom']

'''
Function to extract taxonomic information from all query read placements
and aggregate to form an abundance profile
'''
def queryAbundance(refpkg, query_placement_paths, pool, lock):
    # (1) obtain classification
    _LOG.info("Obtaining read classification from each marker gene")
    classification_paths = {} 

    # failsafe when no placement file are given
    if len(query_placement_paths) == 0:
        _LOG.warning("No placements are found for classification or " \
                "abundance profile, returning...")
        return

    futures = []
    for marker, query_placement_path in query_placement_paths.items():
        clas_outdir = os.path.join(Configs.outdir, 'query_classifications',
                marker)
        if not os.path.isdir(clas_outdir):
            os.makedirs(clas_outdir)

        # necessary files
        taxonomy_path = refpkg[marker]['taxonomy']
        # temp fix mapping issue with taxonomy 
        if not os.path.exists(taxonomy_path):
            taxonomy_path = os.path.join(os.path.dirname(taxonomy_path),
                    'all_taxon.taxonomy')
        mapping_path = refpkg[marker]['seq-to-taxid-map']
        classification_path = os.path.join(clas_outdir, 'placement.classification')
        reordered_placement_path = os.path.join(clas_outdir,
                'placement.reordered.jplace')
        futures.append(pool.submit(getClassification, marker, taxonomy_path,
            mapping_path, query_placement_path, reordered_placement_path,
            clas_outdir, classification_path, lock))

    for future in concurrent.futures.as_completed(futures):
        marker, classification_path = future.result()
        _LOG.info(f"Classification completed on {marker}")
        classification_paths[marker] = classification_path

    # (2) filter classification based on given support value
    support_value = '0.95'
    try:
        support_value = str(
                getattr(Configs, Configs.placement_method).support_value)
    except (AttributeError, ValueError) as e:
        pass
    _LOG.info(f"Filtering with support value={support_value}")

    filtered_paths = {}
    for marker, classification_path in classification_paths.items():
        clas_outdir = os.path.join(Configs.outdir, 'query_classifications',
                marker)
        taxonomy_path = refpkg[marker]['taxonomy']
        if not os.path.exists(taxonomy_path):
            taxonomy_path = os.path.join(os.path.dirname(taxonomy_path),
                    'all_taxon.taxonomy')
        filtered_path = os.path.join(clas_outdir,
                f"placement.classification.{support_value.split('.')[-1]}")
        filterClassification(taxonomy_path,
                classification_path, filtered_path, float(support_value))
        filtered_paths[marker] = filtered_path

    # Updated @ 1.22.2025 - Chengze Shen
    #   - aggregate all classifications and output to a file as well
    # (2.5) aggregate all classifications
    all_classification_path = os.path.join(Configs.outdir,
            'query_classifications.tsv')
    header = False
    with open(all_classification_path, 'w') as f:
        for marker, filtered_path in filtered_paths.items():
            with open(filtered_path, 'r') as fptr:
                lines = fptr.read().strip().split('\n')
            if not header:
                f.write(lines[0] + '\n')
                header = True
            f.write('\n'.join(lines[1:]) + '\n')

    # (3) aggregate classifications to form an abundance profile
    abundance_profile = getAbundanceProfile(refpkg, filtered_paths)

'''
Function to parse filtered classification files and aggregate their abundances
to produce the abundance profile on all taxonomic levels
'''
def getAbundanceProfile(refpkg, filtered_paths):
    global ranks
    _LOG.info("Aggregating abundances for a profile")

    # initializing abundance profile at each taxonomic level
    abundance_profile = {}
    for rank in ranks:
        abundance_profile[rank] = defaultdict(float)
    
    # read in taxid_map from refpkg
    taxid_map = parseTaxonomy(refpkg['taxonomy']['taxonomy'])

    # aggregating results from all markers
    for marker, filtered_path in filtered_paths.items():
        updateAbundanceProfile(filtered_path, abundance_profile)

    # compute percentage at each rank
    _LOG.info("Writing abundance profiles at each taxonomic level to "
            f"{Configs.outdir}") 
    for rank in ranks:
        rank_sum = sum(abundance_profile[rank].values())
        for taxid in abundance_profile[rank].keys():
            abundance_profile[rank][taxid] /= rank_sum

        # write to Configs.outdir as abundance.<rank>.tsv
        outpath = os.path.join(Configs.outdir, f"abundance.{rank}.tsv")
        slist = sorted([(k, v) for k, v in abundance_profile[rank].items()],
                key=lambda x: x[1], reverse=True)
        with open(outpath, 'w') as f:
            f.write('taxa\ttaxid\tabundance\n')
            for (taxid, abundance) in slist:
                # unclassified
                if taxid == 0:
                    f.write('unclassified\t0\t{}\n'.format(abundance))
                else:
                    taxname = taxid_map[taxid][0]
                    f.write('{}\t{}\t{}\n'.format(taxname, taxid, abundance))
        _LOG.info(f"Finished written rank: {rank}") 

'''
Function to update a given abundance profile with the new results
read from the given inpath.
NOTE: updates are "counts", proportions will be calculated later
'''
def updateAbundanceProfile(inpath, abundance_profile):
    global ranks
    with open(inpath, 'r') as f:
        lines = f.read().strip().split('\n')
        # skip header
        for i in range(1, len(lines)):
            parts = lines[i].split('\t')
            taxon = parts[0]
            # fields 1-7 should be what we want
            fields = [int(x) if x != 'NA' else 0 for x in parts[1:]]
            # not classified at all, ignore
            if sum(fields) == 0:
                continue
            
            for j in range(len(ranks)):
                abundance_profile[ranks[j]][fields[j]] += 1

'''
Function to parse a given taxonomy file (all_taxon.taxonomy)
'''
def parseTaxonomy(taxonomy_path):
    taxid_map = {}
    with open(taxonomy_path, 'r') as f:
        for line in f:
            # skip header
            if line.startswith('tax_id'):
                continue
            parts = line.strip().split(',')
            taxid, rank, parent_id, taxname = \
                int(parts[0]), parts[2], int(parts[1]), parts[3]
            taxid_map[taxid] = (taxname, parent_id, rank)
    return taxid_map

'''
Function to reorder a .jplace file to a standard format for reading
'''
def reorderJson(inpath, outpath):
    fh = open(inpath, 'r')
    s_tree = ""
    obj = json.load(fh)
    newobj = {}

    # obtain old order
    old_order = obj['fields']

    new_plc_order = ['edge_num', 'likelihood', 'like_weight_ratio',
            'distal_length', 'pendant_length']
    
    # (1) tree
    newobj['tree'] = obj['tree']
    s_tree = obj['tree'].strip()

    # (2) placements in different order
    newobj['placements'] = [] 
    for placement in obj['placements']:
        if 'nm' in placement:
            ori_p, ori_n = placement['p'], placement['nm']
            ori_n = [x[0] for x in ori_n]
        else:
            ori_p, ori_n = placement['p'], placement['n']
        new_p = []
        for pp in ori_p:
            #new_p.append([pp[1], pp[3], pp[2], pp[0], pp[-1]])
            pp_map = {old_order[i]: pp[i] for i in range(len(pp))}
            tmp = [pp_map[new_plc_order[i]] for i in range(len(new_plc_order))]
            new_p.append(tmp)
        _dict = {'p': new_p, 'n': ori_n}
        newobj['placements'].append(_dict)
    # (3) remaining fields 
    newobj['metadata'] = obj['metadata']
    newobj['version'] = obj['version']
    newobj['fields'] = new_plc_order

    # write to outpath
    ofh = open(outpath, 'w')
    json.dump(newobj, ofh, indent = 4)
    ofh.close()
    fh.close()

    # create stdindata
    s_tree = s_tree.replace("{", "[")
    s_tree = s_tree.replace("}", "]")
    mergeinput = [s_tree]
    mergeinput.append('{}\n{}'.format(s_tree, outpath))
    mergeinput.append("")
    mergeinput.append("")
    s_mergeinput = '\n'.join(mergeinput)
    return s_mergeinput

'''
Function to obtain the taxonomic classification based on placement result,
for the given marker gene taxonomic tree
'''
def getClassification(marker, taxonomy_path, mapping_path,
        ori_jplace_path, jplace_path,
        outdir, classification_path, lock):
    # reorder output .jplace file from the placement method
    stdindata = reorderJson(ori_jplace_path, jplace_path)
    
    # run TIPPJsonMerger software; skip if classification is already there
    if not (os.path.exists(classification_path) \
            and os.stat(classification_path).st_size > 0):
        tippjsonmerger_job = TIPPJsonMergerJob(path=Configs.tippjsonmerger_path,
                taxonomy_path=taxonomy_path, mapping_path=mapping_path,
                outdir=outdir, classification_path=classification_path)
        _ = tippjsonmerger_job.run(stdin=stdindata)

    return marker, classification_path

'''
Function to load in taxonomy
'''
def load_taxonomy(taxonomy_file, lower=True):
    global ranks
    f = open(taxonomy_file, 'r')

    # First line is the keywords for the taxonomy, need to map the keyword to
    # the positional index of each keyword
    results = f.readline().lower().replace('"', '').strip().split(',')
    key_map = dict([(results[i], i) for i in range(0, len(results))])

    # Now fill up taxonomy, level maps keep track of what taxa exist at each
    # level, taxon_map keep track of entire taxonomy
    taxon_map = {}
    level_map = {"species": {}, "genus": {}, "family": {}, "order": {},
            "class": {}, "phylum": {}, "superkingdom": {}}

    for line in f:
        results = line.replace('"', '').strip()
        if (lower):
            results.lower()
        results = results.split(',')
        # insert into taxon map
        taxon_map[results[0]] = results

        # insert into level map
        for level in ranks:
            if (results[key_map[level]] == ''):
                continue
            else:
                if (results[key_map[level]] not in level_map[level]):
                    level_map[level][results[key_map[level]]] = {}
                level_map[level][results[key_map[level]]][results[0]] = \
                    results[0]
    return (taxon_map, level_map, key_map)

'''
Function to filter classification output from a marker gene by a given
support value
'''
def filterClassification(taxonomy_path, classification_path, filtered_path,
        threshold):
    global ranks
    taxon_map, level_map, key_map = load_taxonomy(taxonomy_path)
    
    class_in = open(classification_path, 'r')
    level_map_hierarchy = {"species": 0, "genus": 1, "family": 2, "order": 3,
                "class": 4, "phylum": 5, "superkingdom": 6,
                "root": 7}
    # Need to keep track of last line so we can determine when we switch to
    # new classification
    old_name, old_id, old_rank, old_probability = "", "", "", 1

    # keep track of all fragment names
    names = {}
    classification = {}
    for line in class_in:
        results = line.strip().split(',')
        if (len(results) > 5):
            results = [results[0], results[1], results[2],
                       results[-2], results[-1]]
        (name, id, taxname, rank, probability) = (
            results[0], results[1], results[2], results[3], float(results[4]))
        names[name] = name
        if (name != old_name):
            # when we switch to new fragment, output last classification for
            # old fragment
            if (old_name != ""):
                lineage = taxon_map[old_id]
                output_line = [old_name]
                for level in ranks:
                    clade = lineage[key_map[level]]
                    if (clade == ""):
                        clade = "NA"
                    output_line.append(clade)
                classification[old_name] = output_line
            old_name = name
            old_rank = "root"
            old_probability = 1
            old_id = '1'

        # Switch to new rank if the new probability is higher than threshold
        # and our rank is more specific than our original rank
        if (
            rank in level_map_hierarchy and
            (level_map_hierarchy[old_rank] > level_map_hierarchy[rank]) and
            (probability > threshold)
           ):
            old_rank = rank
            old_probability = probability
            old_id = id
        # Switch to new rank if the new rank matches old rank but has higher
        # probability
        elif (
              rank in level_map_hierarchy and
              (level_map_hierarchy[old_rank] == level_map_hierarchy[rank]) and
              (probability > old_probability)
             ):
            old_rank = rank
            old_probability = probability
            old_id = id

    if old_id in taxon_map:
        lineage = taxon_map[old_id]
        output_line = [old_name]
        for level in ranks:
            clade = lineage[key_map[level]]
            if (clade == ""):
                clade = "NA"
            output_line.append(clade)
        classification[name] = output_line
    class_in.close()

    # write to filtered_path    
    with open(filtered_path, 'w') as class_out:
        class_out.write("fragment\tspecies\tgenus\tfamily\torder\tclass\tphylum\tsuperkingdom\n")
        keys = list(classification.keys())
        keys.sort()
        for frag in keys:
            class_out.write("%s\n" % "\t".join(classification[frag]))



'''
Read binning using BLASTN
'''

import os, time, shutil
from collections import defaultdict

from tipp3 import get_logger
from tipp3.configs import Configs
from tipp3.jobs import BlastnJob
from tipp3.extract_blast_alignment import reverseComplement, \
        extractBlastAlignment

_LOG = get_logger(__name__)

# simple reverse complement map 
#global complement_map
#complement_map = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}

'''
Map reads to refpkg using BLASTN, then process the reads to extract
their assignments and/or alignments to marker genes
'''
def queryBinning(refpkg, query_path):
    database_path = refpkg['blast']['database']
    _LOG.info(f"Initializing BlastnJob with BLASTN database: {database_path}")

    # check if query_path exists
    if not os.path.exists(query_path):
        raise FileNotFoundError(f"{query_path} does not exist!")

    blastn_outdir = os.path.join(Configs.outdir, 'blast_output')
    if not os.path.isdir(blastn_outdir):
        os.makedirs(blastn_outdir)

    # detect previous run, if output exists return
    token_path = os.path.join(blastn_outdir, '_working.token')
    detect_path = os.path.join(blastn_outdir, 'blast.alignment.out')
    if (not os.path.exists(token_path)) and os.path.exists(detect_path):
        _LOG.info(f"Found existing BLAST output: {detect_path}")
        blastn_outpath = detect_path
    else:
        with open(token_path, 'a') as f:
            f.write('placeholder')

        # run BLASTN
        job = BlastnJob(path=Configs.blastn_path,
                query_path=query_path,
                database_path=database_path,
                outdir=blastn_outdir,
                num_threads=Configs.num_cpus)
        blastn_outpath = job.run(logging=False)

        # remove token to indicate job finished
        try:
            os.remove(token_path)
        except FileNotFoundError:
            pass

    # process BLASTN output
    query_blast_paths, query_aln = processBlastnOutput(refpkg, blastn_outpath,
            blastn_outdir)

    # write query reads to local for alignments, truncate if neccessary
    query_outdir = os.path.join(Configs.outdir, 'query')
    query_paths = splitQueries(refpkg, query_aln, query_outdir)

    # if Configs.alignment_method is BLAST, extract alignment at this step
    query_alignment_paths = {}
    if Configs.alignment_method == 'blast':
        query_alignment_paths = extractBlastAlignment(refpkg,
                Configs.outdir, query_blast_paths)

    return query_paths, query_alignment_paths

'''
process the BLASTN output, extract assignment information and/or alignment
'''
def processBlastnOutput(refpkg, blastn_outpath, blastn_outdir):
    # seq-to-marker map
    map_path = refpkg['blast']['seq-to-marker-map']
    gene_mapping = readGeneMapping(map_path)

    # read blastn output to memory
    threshold = 50
    try:
        threshold = int(getattr(Configs, 'blast').threshold)
    except (AttributeError, ValueError) as e:
        pass
    _LOG.info(f"Filtering BLASTN results: >= {threshold}bp coverage.")

    detect_path = blastn_outpath + '.finished'
    if not os.path.exists(detect_path):
        query_aln, query_markers_mapped = readBlastnOutput(
                blastn_outpath, gene_mapping, threshold)
        query_blast_paths = {}

        # write the mapped marker gene(s) for each mapped read
        mapped_path = '{}/query_mapped_markers.tsv'.format(blastn_outdir)
        with open(mapped_path, 'w') as f:
            f.write('read_name\tnum_mapped_markers\tmapped_markers\n')
            for k, v in query_markers_mapped.items():
                f.write('{}\t{}\t{}\n'.format(k, len(v), '|'.join(v)))
        
        # read in marker genes to use
        markers = refpkg['genes']

        # write to blastn_outdir as checkpoints
        write_order = ['source_taxon', 'qstart', 'qend', 'qaln', 'sstart', 'send', 'saln']
        marker_fptr = {}
        for marker in markers:
            _path = '{}/queries-{}.out'.format(blastn_outdir, marker) 
            marker_fptr[marker] = {}
            marker_fptr[marker]['path'] = _path
            marker_fptr[marker]['fptr'] = open(_path, 'w')
            marker_fptr[marker]['count'] = 0
        
        for k, v in query_aln.items():
            # k --> (taxon_name, marker)
            taxon = k[0]
            marker = v['source_taxon'][1]
            if marker in marker_fptr:
                towrite = [v[_x] for _x in write_order]
                marker_fptr[marker]['fptr'].write('>{}\n{}\n'.format(taxon, towrite))
                marker_fptr[marker]['count'] += 1

        for marker in markers:
            marker_fptr[marker]['fptr'].close()
            _path = marker_fptr[marker]['path']
            # remove empty marker gene outputs
            if marker_fptr[marker]['count'] == 0:
                _LOG.debug(f'{marker} has no assigned queries, removing {_path}')
                os.system('rm {}'.format(_path))
            else:
                query_blast_paths[marker] = marker_fptr[marker]['path']

        # write to local to indicate that splitting is done
        with open(detect_path, 'w') as f:
            f.write('placeholder')
    else:
        # read from existing split queries
        _LOG.info(f"Found existing filtered BLASTN results")
        query_blast_paths, query_aln = readSplitQueries(blastn_outdir)
    return query_blast_paths, query_aln


######################## HELPER FUNCTIONS ###########################

#'''
#helper function to obtain the reverse complement of a sequence
#'''
#def reverseComplement(seq):
#    try:
#        char_list = [complement_map[c] for c in seq]
#    except KeyError:
#        char_list = [complement_map[c] if c in complement_map else c for c in seq]
#    new_seq = ''.join(char_list)
#    return new_seq[::-1]

'''
helper function to read marker gene mapping
'''
def readGeneMapping(map_path, delimiter='\t'):
    mapping = {}
    with open(map_path, 'r') as f:
        for line in f:
            parts = line.strip().split(delimiter)
            mapping[parts[0]] = parts
    return mapping

'''
helper function to read in genes files that are already split
NOTE: only when all query reads are split this function will be invoked
'''
def readSplitQueries(blastn_outdir):
    # removing irrelevant files
    existing_files = os.popen(f'ls {blastn_outdir}').read().strip().split('\n')
    existing_files = [x for x in existing_files if x.startswith('queries')]
    _LOG.info(
        f"Existing queries are assigned to ({len(existing_files)}) marker genes.")

    query_blast_paths = {}
    for f in existing_files:
        marker = f.split('-')[1].split('.')[0]
        query_blast_paths[marker] = os.path.join(blastn_outdir, f)

    query_aln = defaultdict(dict)
    for marker, infile in query_blast_paths.items():
        cur_taxon = ""; cur_aln = None
        f = open(infile, 'r')
        for line in f:
            if line.startswith('>'):
                if cur_taxon != "":
                    query_aln[cur_taxon] = cur_aln
                    cur_aln = None
                else:
                    cur_taxon = line.strip()[1:]
            else:
                parts = eval(line.strip())
                cur_aln = {'source_taxon': parts[0],
                        'qcov': parts[2] - parts[1],
                        'qstart': parts[1], 'qend': parts[2],
                        'qaln': parts[3],
                        'sstart': parts[4], 'send': parts[5],
                        'saln': parts[6]}
        if cur_taxon != "":
            query_aln[cur_taxon] = cur_aln
        f.close()
    return query_blast_paths, query_aln

'''
helper function to read in blastn output
'''
def readBlastnOutput(blastn_outpath, gene_mapping, threshold):
    query_aln = defaultdict(dict)
    query_markers_mapped = defaultdict(set)

    with open(blastn_outpath, 'r') as f:
        line = f.readline()
        b_consider, scan_mode, cur_taxon = False, False, ''
        mapped_markers = set()

        # initialization of the following
        # source_taxon, qcov, qstart, qend, qaln, sstart, send, saln
        cur_aln = {'source_taxon': ('', ''), 'qcov': -1,
                'qstart': -1, 'qend': -1, 'qaln': '',
                'sstart': -1, 'send': -1, 'saln': ''}

        while line:
            line = line.split('\n')[0].strip()
            # start of a new query entry, starting reading
            if line.startswith('Query='):
                taxon = line.split('= ')[1]
                b_consider = True
                cur_taxon = taxon
            # reach the end of a query entry
            elif line.startswith('Effective search space'):
                # check if (1) qcov >= threshold
                #          (2) cur_aln has better coverage than previous one
                if cur_taxon != '':
                    cur_aln, _ = updateQueryAlignment(
                            query_aln, cur_taxon, cur_aln, threshold,
                            mapped_markers)
                    if len(mapped_markers) > 0:
                        query_markers_mapped[cur_taxon].update(mapped_markers)
                    cur_taxon = ''
                    scan_mode = False
                    mapped_markers.clear()
            # is actively reading in entries for a query
            elif b_consider:
                if line.startswith('>'):
                    # cur_aln is not empty, flush current results before moving
                    # to the next one
                    if cur_aln['source_taxon'][1] != '':
                        cur_aln, mapped_markers = updateQueryAlignment(
                                query_aln, cur_taxon, cur_aln, threshold,
                                mapped_markers)
                    source = line.split('>')[1]
                    # obtain corresponding marker gene from gene_mapping
                    scan_mode = True
                    cur_aln['source_taxon'] = (source, gene_mapping[source][1])

                    # skip scanning since we just start, move to next line
                    line = f.readline()
                    continue
                
                # is actively scanning for alignment part
                if scan_mode:
                    # query alignment part
                    if line.startswith('Query'):
                        # parts[1]: current qstart
                        # parts[2]: current qaln
                        # parts[3]: current qend
                        parts = line.split()
                        if cur_aln['qstart'] == -1:
                            cur_aln['qstart'] = int(parts[1])
                        cur_aln['qaln'] += parts[2]
                        cur_aln['qend'] = int(parts[3])
                    # target alignment part
                    elif line.startswith('Sbjct'):
                        parts = line.split()
                        if cur_aln['sstart'] == -1:
                            cur_aln['sstart'] = int(parts[1])
                        cur_aln['saln'] += parts[2]
                        cur_aln['send'] = int(parts[3])
            # move to next line
            line = f.readline()
    return query_aln, query_markers_mapped

'''
helper function to update query_aln with a new entry recorded
replace the old one if the new one has longer coverage
returns a completely new entry for the following scan
'''
def updateQueryAlignment(query_aln, cur_taxon, cur_aln, threshold,
        mapped_markers):
    # compute current query coverage
    mapped_marker = cur_aln['source_taxon'][1]
    new_mapped_markers = set(mapped_markers)
    cur_aln['qcov'] = abs(cur_aln['qstart'] - cur_aln['qend']) + 1
    if cur_aln['qcov'] >= threshold:
        # new entry
        # go over previously mapped markers and identify their mapped regions.
        # check if there is any overlap with cur_aln
        can_add = True
        for prev_marker in mapped_markers:
            # ASSUMPTION: prev_marker exists
            prev_aln = query_aln[(cur_taxon, prev_marker)]
            prev_start, prev_end = prev_aln['qstart'], prev_aln['qend']
            prev_cov = prev_aln['qcov']
            # (1) overlaps
            if max(prev_start, cur_aln['qstart']) < min(prev_end, cur_aln['qend']):
                if prev_cov < cur_aln['qcov']:
                    # remove old mapping if its coverage is smaller
                    query_aln.pop((cur_taxon, prev_marker))
                    new_mapped_markers.remove(prev_marker)
                else:
                    # we cannot add cur_aln to query_aln
                    can_add = False
        # (2) either no overlap or overlap wins
        if can_add:
            query_aln[(cur_taxon, mapped_marker)] = cur_aln
            new_mapped_markers.add(mapped_marker)
        #if len(query_aln[cur_taxon]) == 0:
        #    query_aln[cur_taxon] = cur_aln
        #    # compare current qcov to the one already stored
        #elif cur_aln['qcov'] > query_aln[cur_taxon]['qcov']:
        #    query_aln[cur_taxon] = cur_aln
    return {'source_taxon': ('', ''), 'qcov': -1,
        'qstart': -1, 'qend': -1, 'qaln': '',
        'sstart': -1, 'send': -1, 'saln': ''}, new_mapped_markers

'''
helper function to slit queries to marker genes and write to local for
alignment later
'''
def splitQueries(refpkg, query_aln, query_outdir):
    _LOG.info("Writing assigned query reads to local.")
    markers = refpkg['genes']

    if not os.path.isdir(query_outdir):
        os.makedirs(query_outdir)

    ret = {}
    detect_path = os.path.join(query_outdir, 'blast-binned.out')
    if os.path.exists(detect_path):
        _LOG.info("Found assigned query reads, using existing ones...")
        for marker in markers:
            _outpath = os.path.join(query_outdir, '{}.queries.fasta'.format(marker))
            if os.path.exists(_outpath):
                ret[marker] = _outpath
        return ret

    # initialize assigned queries for each marker gene
    marker_fptr = {}
    for marker in markers:
        _outpath = os.path.join(query_outdir, '{}.queries.fasta'.format(marker))
        ret[marker] = _outpath
        marker_fptr[marker] = open(_outpath, 'w')

    # write blastn alignment output info
    info_f = open(detect_path, 'w')
    info_f.write('qseqid,sseqid,marker,trim_qstart,trim_qend,qlen\n')

    # go over each entry and write query to corresponding marker gene output
    # ONLY CONSIDER THE MARKERS THAT ARE USED
    for k, v in query_aln.items():
        # k --> (taxon_name, marker)
        taxon = k[0]
        sseqid, smarker = v['source_taxon']
        qstart, qend, sstart, send = v['qstart'], v['qend'], v['sstart'], v['send']
        if smarker not in marker_fptr:
            continue

        # retain only the part that's mapped by BLASTN
        seq = v['qaln'].replace('-', '').upper()

        # reverse complement
        if sstart > send:
            rev_seq = reverseComplement(seq)
        else:
            rev_seq = seq
        
        # write to corresponding marker gene (if using)
        marker_fptr[smarker].write('>{}\n{}\n'.format(taxon, rev_seq))
        info_f.write('{},{},{},{},{},{}\n'.format(taxon, sseqid, smarker,
            qstart, qend, qend - qstart + 1))

    for marker in markers:
        marker_fptr[marker].close()
    info_f.close()

    # remove empty queries.fasta
    tmp_markers = list(ret.keys())
    for marker in tmp_markers:
        _outpath = ret[marker]
        if os.stat(_outpath).st_size == 0:
            _LOG.debug(f'Removing redundant {marker}: {_outpath}')
            os.remove(_outpath)
            ret.pop(marker)

    _LOG.info(f"Inputs mapped to ({len(ret)}) marker genes: {ret.keys()}")
    _LOG.info("Number of mapped reads (one read may map to " \
            f"multiple marker genes): {len(query_aln)}")
    return ret




'''
Aligning query reads to their assigned marker genes, using either BLASTN
output or some other alignment method
'''

import os, time, shutil
from tipp3 import get_logger
from tipp3.configs import Configs
from tipp3.helpers.alignment_tools import Alignment
from tipp3.jobs import WITCHAlignmentJob

_LOG = get_logger(__name__)

'''
Align query reads to marker genes with the user-defined alignment methods
'''
def queryAlignment(refpkg, query_paths):
    # query_alignments only retain query alignments
    # the backbone part is unnecessary
    _LOG.info(f"Started alignment method: {Configs.alignment_method}")

    query_alignment_paths = {}
    additional_kwargs = {}
    try:
        additional_kwargs = getattr(Configs, Configs.alignment_method).__dict__
    except (AttributeError, ValueError) as e:
        pass
    _LOG.info("(From main.config) parameter options for " \
            f"{Configs.alignment_method}: {additional_kwargs}")

    # removing temporary files that we do not care
    temp_files = ['est.aln.fasta', 'tree_decomp', 'debug.txt']

    idx = 0; total = len(query_paths)
    for marker, query_path in query_paths.items():
        _LOG.info(f"Aligning query reads for {marker}")
        idx += 1
        alignment_dir = os.path.join(Configs.outdir, 'query_alignments',
                marker)
        if not os.path.isdir(alignment_dir):
            os.makedirs(alignment_dir)

        # multi-level detect to avoid rerunning:
        # first detect if 'est.aln.fasta' exists
        d1_path = os.path.join(alignment_dir, 'est.aln.masked.fasta')
        d2_path = os.path.join(alignment_dir, 'est.aln.masked.fasta.gz')
        skip_alignment = False
        if os.path.exists(d1_path) and os.stat(d1_path).st_size > 0:
            # gzip the file (if successful, will remove d1_path)
            _LOG.info(f"Found existing alignment: {d1_path}, compressing...")
            #os.system(f"lz4 --rm -f -q {d1_path} {d2_path}")
            os.system(f"gzip {d1_path}")
            skip_alignment = True
        elif os.path.exists(d2_path) and os.stat(d2_path).st_size > 0:
            # keep the gzip file and avoid re-running the alignment step
            _LOG.info(f"Found existing compressed alignment: {d2_path}")
            skip_alignment = True

        # not skipping alignment and need to generate the alignment file
        # to extract query alignments
        if not skip_alignment:
            # obtain the backbone alignment and tree paths for alignment
            backbone_path = refpkg[marker]['alignment']
            backbone_tree_path = refpkg[marker]['alignment-decomposition-tree']

            # identify job type - different alignment method(s)
            if Configs.alignment_method == 'witch':
                alignment_job = WITCHAlignmentJob(path=Configs.witch_path,
                        query_path=query_path,
                        backbone_path=backbone_path,
                        backbone_tree_path=backbone_tree_path,
                        outdir=alignment_dir,
                        num_cpus=Configs.num_cpus,
                        **additional_kwargs)
            else:
                raise NotImplementedError(
                        f"Alignment method {Configs.alignment_method} " \
                        " is not implemented yet.")
            # gzip the output masked file
            # raw_alignment_path == d1_path
            raw_alignment_path = alignment_job.run(logging=True)
            os.system(f"gzip {raw_alignment_path}")
            #os.system(f"lz4 --rm -f -q {raw_alignment_path} {d2_path}")

        # extract query alignment from {d2_path}
        queryunaln = Alignment(); queryunaln.read_file_object(query_path)
        query_keys = list(queryunaln.keys())
        aln = Alignment(); aln.read_file_object(d2_path)
        subaln = aln.sub_alignment(query_keys)
        # write to local
        query_alignment_path = os.path.join(alignment_dir,
                'est.aln.masked.queries.fasta')
        subaln.write(query_alignment_path, 'FASTA')
        query_alignment_paths[marker] = query_alignment_path

        # clean up some temp files
        for tf in temp_files:
            toremove = os.path.join(alignment_dir, tf)
            if not os.path.exists(toremove):
                continue
            if os.path.isfile(toremove):
                os.remove(toremove)
            else:
                shutil.rmtree(toremove)
        _LOG.info(f"({idx}/{total}) Finished aligning query reads for {marker}")
    return query_alignment_paths



import time
import numpy as np
import re, os, sys
import itertools
from copy import deepcopy
from random import random
from functools import reduce
from dendropy.datamodel.taxonmodel import Taxon
import time
import io
import gzip
from collections import defaultdict
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping
from operator import add

try:
    filetypes = (io.IOBase, file)
except NameError:
    filetypes = io.IOBase

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
    from io import BytesIO

gappat = re.compile(r'[-?]+')
nogappat = re.compile(r'[^-?]+')
_INDEL = re.compile(r"[-]")
_COACH_INS = re.compile(r"[.]")
_DANGEROUS_NAME_CHARS = re.compile(r"[^a-zA-Z0-9]")
ILLEGAL_CHARS = re.compile(r"[^a-zA-Z?-]")

################    Adapted from SEPP.filemgr.py    ##################
def open_with_intermediates(filepath, mode):
    d = os.path.dirname(filepath)
    if d:
        if not os.path.exists(d):
            os.makedirs(d)
        elif not os.path.isdir(d):
            raise IOError(
                "The file %s cannot be created because %s exists "
                "but is not a directory" % (filepath, d))
    return open(filepath, mode)

################ Copied from PASTA-Alignment        ##################
################ CompactAlignment class definition  ##################
def write_compact_to_fasta(alignment, dest):
    """Writes the `alignment` in FASTA format to either a file object or file"""
    file_obj = None
    if isinstance(dest, str):
        file_obj = open(dest, "w")
    else:
        file_obj = dest
    for name in list(alignment.keys()):
        s = alignment.as_string_sequence(name)
        file_obj.write('>{}\n{}\n'.format(name, s))
    if isinstance(dest, str):
        file_obj.close()

def read_fasta(src, remove_gaps=False):
    """generator that returns (name, sequence) tuples from either a FASTA
    formatted file or file object.
    """
    file_obj = None
    if isinstance(src, str):
        # allow opening with '.gz' suffix file
        if src.split('.')[-1] == 'gz':
            try:
                file_obj = gzip.open(src, 'rt')
            except IOError:
                print(("The file `%s` does not exist, exiting gracefully" % src))
        else:
            try:
                file_obj = open(src, "r")
            except IOError:
                print(("The file `%s` does not exist, exiting gracefully" % src))
    elif isinstance(src, filetypes):
            file_obj = src
    else:
        raise TypeError('FASTA reader cannot recognize the source of %s, %s, %s' % (src,type(src),isinstance(src, filetypes)))
    name = None
    seq_list = list()
    for line_number, i in enumerate(file_obj):
        if i.startswith('>'):
            if name:
                if remove_gaps:
                    yield name, ''.join(seq_list).replace('-', '')
                else:
                    yield name, ''.join(seq_list)
                seq_list = list()
            name = i[1:].strip()
        else:
            #seq = ''.join(i.strip().upper().split())
            seq = ''.join(i.strip().split())
            #if not is_sequence_legal(seq):
            #    raise Exception("Error: illegal characeters in sequence at line %d" % line_number)
            seq_list.append(seq)
    if name:
        if remove_gaps:
            yield name, ''.join(seq_list).replace('-', '')
        else:
            yield name, ''.join(seq_list)
    if isinstance(src, str):
        file_obj.close()

def write_fasta(alignment, dest):
    """Writes the `alignment` in FASTA format to either a file object or file"""
    file_obj = None
    if isinstance(dest, str):
        file_obj = open(dest, "w")
    else:
        file_obj = dest
    for name, seq in list(alignment.items()):
        file_obj.write('>%s\n%s\n' % (name, seq) )
    if isinstance(dest, str):
        file_obj.close()

'''
Infer data type from file
'''
def inferDataType(path):
    sequences = read_fasta(path, remove_gaps=True)
    acg, t, u, total = 0, 0, 0, 0
    for taxon, seq in sequences:
        letters = seq.upper()
        for letter in letters:
            total = total + 1
            
            if letter in ('A', 'C', 'G', 'N'):
                acg = acg + 1
            elif letter == 'T':
                t = t + 1
            elif letter == 'U':
                u = u + 1
    
    if u == 0 and (acg + t)/total > 0.9:
        #print("Found {}% ACGT-N, assuming DNA..".format(int(100*(acg + t)/total)))
        dataType = "dna"
    elif t == 0 and (acg + u)/total > 0.9:
        #print("Found {}% ACGU-N, assuming RNA..".format(int(100*(acg + u)/total)))
        dataType = "rna"
    else:
        #print("Assuming protein..")
        dataType = "amino"
          
    return dataType


class Alignment(dict, object):
    """A simple class that maps taxa names to sequences.
    TODO: switch to dendropy character_matrix
    """
    def __init__(self):
        "creates an empty matrix"
        dict.__init__(self)
        self.datatype = None

    def get_datatype(self):
        return self._datatype

    def set_datatype(self, d):
        if d is None:
            self._datatype = None
        else:
            self._datatype = d.upper()

    datatype = property(get_datatype, set_datatype)

    def get_sequence_names(self):
        "returns a list of sequence names"
        return list(self.keys())

    def get_num_taxa(self):
        "returns the number sequences"
        return len(self.get_sequence_names())

    def read_file_object(self, file_obj, file_format='FASTA'):
        """Augments the matrix by reading the file object.
        If duplicate sequence names are encountered then the old name will be replaced.
        """
        if ( file_format.upper() == 'FASTA' ):
            read_func = read_fasta
        #elif ( file_format.upper() == 'NEXUS' ):
        #    read_func = read_nexus
        #elif ( file_format.upper() == 'PHYLIP' ):
        #    read_func = read_phylip
        #elif ( file_format.upper() == 'COMPACT3' ):
        #    read_func = read_compact3
        else:
            raise NotImplementedError("Unknown file format (%s) is not supported" % file_format)
        for name, seq in read_func(file_obj):
            self[name] = seq

    def write_filepath(self, filename, file_format='FASTA', zipout=False):
        """Writes the sequence data in the specified `file_format` to `filename`"""
        
        file_obj = open_with_intermediates(filename,'w')
        if zipout:
            file_obj.close()
            file_obj = StringIO()
        self.write(file_obj, file_format=file_format)
        if zipout:
            import gzip
            file_obj_gz = gzip.open(filename, "wb", 6)
            file_obj_gz.write(str.encode(file_obj.getvalue()))
            file_obj_gz.close()
        file_obj.close()

    def write(self, file_obj, file_format):
        """Writes the sequence data in the specified `file_format` to `file_obj`"""
        if ( file_format.upper() == 'FASTA' ):
            write_func = write_fasta
        #elif ( file_format.upper() == 'NEXUS' ):
        #    write_func = write_nexus
        #elif ( file_format.upper() == 'PHYLIP' ):
        #    write_func = write_phylip
        #elif ( file_format.upper() == 'COMPACT' ):
        #    write_func = write_compact            
        #elif ( file_format.upper() == 'COMPACT2' ):
        #    write_func = write_compact2       
        #elif ( file_format.upper() == 'COMPACT3' ):
        #    write_func = write_compact3
        else:
            write_func = write_fasta
        write_func(self, file_obj)

    def unaligned(self):
        """
        Returns a new alignment with all gaps and missing sequences removed.
        """
        new_alignment = Alignment()
        new_alignment.datatype = self.datatype
        for name, seq in self.items():
            new_seq = re.sub(_INDEL, '', str(seq))
            if new_seq != '':
                new_alignment[name] = new_seq
        return new_alignment

    def sub_alignment(self, sub_keys):
        "Creates an new alignment with a subset of the taxa."
        new_alignment = Alignment()
        new_alignment.datatype = self.datatype
        for key in sub_keys:
            if key in self:
                new_alignment[key] = self[key]
        return new_alignment

    def is_empty(self):
        return self.__len__() < 1

    def is_aligned(self):
        if self.is_empty():
            raise ValueError("The alignment is empty.\n")
        else:
            v = list(self.values())
            first_seq_len = len(v[0])
            return all([len(i) == first_seq_len for i in v])

    def partition_info(self, base=0):
        return (self.datatype, 1+base, self.sequence_length() + base)

    def sequence_length(self):
        if self.is_aligned():
            return len(list(self.values())[0])

    def max_sequence_length(self):
        return max(len(re.sub(_INDEL, '', v)) for v in list(self.values()))

    def get_all_gap_cols(self):
        all_gaps = list(range(0, self.sequence_length()))
        for seq in self.values():
            all_gaps[:] = [i for i in all_gaps if seq[i] == '-']
        return all_gaps

    def remove_columns(self, indexes):
        for name, seq in self.items():
            self[name] = ''.join((char for idx, char in enumerate(seq)
                if idx not in indexes))

    def delete_all_gaps(self):
        rem = set(self.get_all_gap_cols())
        subset = [x for x in range(0, self.sequence_length()) if x not in rem]
        self.remove_columns(set(rem))
        return subset
    
    def from_bytearray_to_string(self):
        for k,v in self.items():
            self[k] = str(v)

    def from_string_to_bytearray(self):
        for k,v in self.items():
            self[k] = bytearray(v)   
    
    def divide_to_equal_chunks(self, chunks, max_chunk_size=None):
        names = self.get_sequence_names()
        ret = []
        if max_chunk_size and len(names) / chunks > max_chunk_size:
            chunks = len(names) // max_chunk_size + 1
        for i in range(0, chunks):
            subset = names[i:len(names):chunks]
            if subset:
                subset_alg = self.sub_alignment(subset)
            else:
                subset_alg = None
            ret.append(subset_alg)
        return ret

    def mask_gapy_sites(self,minimum_seq_requirement):        
        n = len(list(self.values())[0])
        # The following implements column-based masking. Seems to be more efficient than row based
        masked = []
        allseqs = list(self.values())
        allseqs.sort(key=lambda x: x.count("-"))
        for c in range(0,n):
            r = minimum_seq_requirement
            for seq in allseqs:
                if seq[c] != "-":
                    r -= 1
                if r <= 0:
                    break
            if r > 0:
                masked.append(c)
                 
        #_LOG.debug("%d Columns identified for masking" %len(masked))
        if not masked:
            return
        included = [z for z in reduce(lambda x,y: x+[(x[-1][1]+1,y)],masked,[(-1,-1)]) if z[0]!=z[1]]
        if not included:
            included.append((masked[-1]+1,n))
        if included[-1][1] < n and masked[-1]+1 != n:
            included.append((masked[-1]+1,n))
        for k,seq in self.items():
            tmp = []
            for (i,j) in included:
                tmp.append(seq[i:j])
            self[k] = "".join(tmp)
        nn = len(list(self.values())[0])
        assert (len(masked) == n - nn), "Masking results is not making sense: %d %d %d" %(len(masked), n , nn)
        #_LOG.debug("Masking done. Before masking: %d; After masking: %d; minimum requirement: %d;" %(n,nn,minimum_seq_requirement))

    def merge_in(self, she):
        pass
        # NOT IMPLEMENTED
        #merge_in(self,she)


class AlignmentSequence:
    def __init__(self):
        self.seq = None
        self.pos = []
    
    def replace(self, match_char, replace_char):
        s = AlignmentSequence()
        s.pos.extend(self.pos)
        s.seq = self.seq.replace(match_char, replace_char)
        return s
    
    def as_string(self, pad_to):
        s=[]
        nxt = 0
        for i,p in enumerate(self.pos):
            if nxt != p:
                s.append("-" * (p-nxt))
            s.append(self.seq[i]) 
            nxt = p+1
        if (pad_to > nxt):
            s.append("-" * (pad_to-nxt))
        return "".join(s)
    
    def __str__(self):
        return self.as_string(0)
    
    def __repr__(self):
        return self.__str__()

class CompactAlignment(dict,object):
    def __init__(self):
        self.colcount = 0
        self.datatype = None
    
    def sub_alignment(self, sub_keys):
        "Creates an new alignment with a subset of the taxa."
        new_alignment = CompactAlignment()
        new_alignment.datatype = self.datatype
        for key in sub_keys:
            if key in self:
                new_alignment[key] = self[key]
        return new_alignment
   
    def is_aligned(self):
        return True
    def sequence_length(self):
        return self.colcount
    def get_num_taxa(self):
        return len(self)
    
    #def unaligned(self):
    #    n = Alignment()
    #    n.datatype = self.datatype
    #    for k,seq in self.items():
    #        n[k] = seq.seq
    #    return n
    
    def iter_column_character_count(self, seqsubset = None):
        if seqsubset is None:
            seqsubset = list(self.keys())
        
        counts = [0] * self.colcount
        for k in seqsubset:
            for pos in self[k].pos:
                counts[pos] += 1
        
        for c in counts:
            yield c
            
    def iter_columns_with_minimum_char_count(self, x, seqsubset = None):
        for i,c in enumerate(self.iter_column_character_count(seqsubset)):
            if c >= x:
                yield i

    def iter_columns_with_maximum_char_count(self, x, seqsubset = None):
        for i,c in enumerate(self.iter_column_character_count(seqsubset)):
            if c <= x:
                yield i
            
                
    def get_insertion_columns(self,shared):
        return set(i for i in self.iter_columns_with_maximum_char_count(0,shared)) 

    def merge_in(self, she):
        '''
        Merges she inside self, assuming we share some common taxa, and the 
        alignment of common taxa is identical across both alignments.
        
        When assumptions are not met, behavior is largely undefined. 
        '''      
        #global _T_ID
        #_T_ID += 1
        #ID = _T_ID    
        #TIMING_LOG.info("transitivitymerge (%d) started" %ID )    
        mykeys = set(self.keys())
        herkeys = set(she.keys())
        #_LOG.debug("Transitive Merge Started. ID:%d - Rows: %d,%d" %(ID,len(mykeys),len(herkeys)))    
        shared = mykeys.intersection(herkeys)
        #_LOG.debug("Shared seq: %d" %(len(shared)))        
        onlyhers = herkeys - shared
        me_ins = self.get_insertion_columns(shared)
        she_ins = she.get_insertion_columns(shared)
        #_LOG.debug("Insertion Columns: %d,%d" %(len(me_ins),len(she_ins)))
        
        memap=[]
        shemap=[]
        
        melen = self.colcount
        shelen =  she.colcount
    
        ime = 0
        ishe = 0
        inew = 0
        while ime < melen or ishe < shelen:
            if ime in me_ins:              
                memap.append(inew)
                ime += 1
            elif ishe in she_ins:
                shemap.append(inew)
                ishe += 1
            else:       
                memap.append(inew)
                shemap.append(inew)
                ishe += 1
                ime += 1
            inew += 1
            
        self.colcount = inew
        
        for seq in self.values():
            seq.pos = [memap[p] for p in seq.pos]
            
        for k in onlyhers:
            she[k].pos = [shemap[p] for p in she[k].pos]
            self[k] = she[k]        
            
        #TIMING_LOG.info("transitivitymerge (%d) finished" %ID )
        #_LOG.debug("Transitive Merge Finished. ID:%d; cols after: %d" %(ID,self.colcount))

    def mask_gapy_sites(self,minimum_seq_requirement):                
        #_LOG.debug("Masking alignment sites with fewer than %d characters from alignment with %d columns" 
        #           %(minimum_seq_requirement,self.colcount))

        masked = set(self.iter_columns_with_maximum_char_count(minimum_seq_requirement-1))
        
        #_LOG.debug("%d Columns identified for masking" %len(masked))
        
        self.mask_sites(masked)
        
    def mask_unaligned_sites(self):
        #_LOG.debug("Masking alignment sites with lower case letters from an alignment with %d sites" 
        #           %(self.colcount))

        masked = set()
        for seq in self.values():
            for c,i in zip(seq.seq,seq.pos):
                if c > 'a' and c < 'z':
                    masked.add(i)
        
        #_LOG.debug("%d Columns identified for masking" %len(masked))
        
        self.mask_sites(masked)     
        
        return self   
        
    def mask_sites(self, masked):
        if not masked:
            return
        
        off = 0
        colmap = []
        for c in range(0,self.colcount):
            if c in masked:
                off += 1
                colmap.append(-1)
            else:
                colmap.append(c - off)
                
        #_LOG.debug("Column index mapping calculated.")
        for seq in self.values():
            # Find ID of char positions that would be masked
            maskind = [i for i,c in enumerate(seq.pos) if c in masked]
            n = len(seq.pos)
            included = [z for z in reduce(lambda x,y: x+[(x[-1][1]+1,y)],maskind,[(-1,-1)]) if z[0]!=z[1]]
            if not maskind:
                included.append((0,n))
            elif (not included or included[-1][1] < n) and maskind[-1]+1 != n:
                included.append((maskind[-1]+1,n))
            tmp = []
            for (i,j) in included:
                tmp.extend(seq.seq[i:j])
            seq.seq = "".join(tmp)
            seq.pos = [colmap[x] for x in (p for p in seq.pos if p not in masked)]
                
        self.colcount -= off
        #_LOG.debug("Masking done. Before masking: %d; After masking: %d;" 
        #           %(self.colcount+off,self.colcount))
        
    def read_filepath(self, filename, file_format='FASTA'):
        """Augments the matrix by reading the filepath.
        If duplicate sequence names are encountered then the old name will be replaced.
        """
        file_obj = open(filename, 'r')
        ret = self.read_file_object(file_obj, file_format=file_format)
        file_obj.close()
        return ret

    def read_file_object(self, file_obj, file_format='FASTA'):
        """Augments the matrix by reading the file object.
        If duplicate sequence names are encountered then the old name will be replaced.
        """
        if ( file_format.upper() == 'FASTA' ):
            read_func = read_fasta        
        #elif ( file_format.upper() == 'COMPACT' ):
        #    read_func = read_compact
        #elif ( file_format.upper() == 'COMPACT3' ):
        #    read_func = read_compact3
        else:
            raise NotImplementedError("Unknown file format (%s) is not supported" % file_format)
        self.colcount = 0
        for name, seq in read_func(file_obj):
            cseq, l = self.get_alignment_seq_object(seq)
            self[name] = cseq
            self.colcount = max(l, self.colcount)
        
    def as_string_sequence(self,name):
        seq = self[name]
        return seq.as_string(self.colcount)
            
    def get_alignment_seq_object(self, seq):
        cseq = AlignmentSequence()
        l = 0
        if isinstance(seq, str):            
            for m in re.finditer(nogappat, seq):
                cseq.pos.extend(range(m.start(),m.end()))
            cseq.seq = re.sub(gappat,"",seq)
            l = len(seq)
        else:
            cseq.seq = seq[0]
            cseq.pos = seq[1]
            l = seq[1][-1] + 1
        return (cseq,l)

    def update_dict_from(self, alignment):
        for k in self.keys():
            alignment[k] = self.as_string_sequence(k)
        alignment.datatype = self.datatype
            
    def update_from_alignment(self, alignment):
        for k,v in alignment.items():
            self[k],l = self.get_alignment_seq_object(v)
        self.colcount = l
        self.datatype = alignment.datatype
            
    def write_filepath(self, filename, file_format='FASTA', zipout=False):
        """Writes the sequence data in the specified `file_format` to `filename`"""
        
        file_obj = open_with_intermediates(filename,'w')
        if zipout:
            file_obj.close()
            file_obj = StringIO()
        self.write(file_obj, file_format=file_format)
        if zipout:
            import gzip
            file_obj_gz = gzip.open(filename, "wb", 6)
            file_obj_gz.write(str.encode(file_obj.getvalue()))
            file_obj_gz.close()
        file_obj.close()

    def write(self, file_obj, file_format):
        """Writes the sequence data in the specified `file_format` to `file_obj`"""
        if ( file_format.upper() == 'FASTA' ):
            write_func = write_compact_to_fasta        
        #elif ( file_format.upper() == 'COMPACT' ):
        #    write_func = write_compact_to_compact
        #elif ( file_format.upper() == 'COMPACT3' ):
        #    write_func = write_compact_to_compact3 
        #elif ( file_format.upper() == 'PHYLIP' ):
        #    write_func = write_compact_to_phylip 
        else:
            write_func = write_compact_to_fasta
        write_func(self, file_obj)

def compact(alg):
    comp = CompactAlignment()
    comp.update_from_alignment(alg)
    return comp

################# Adapted from SEPP/alignment.py ################
class ReadOnlyAlignment(Mapping, object):
    def get_num_taxa(self):
        return len(self.get_sequence_names())
    
    def get_length(self):
        if self.is_empty():
            raise ValueError("The alignment is empty.\n")
        else:
            return len(next(iter(self.values())))

    def get_sequence_names(self):
        return list(self.keys())

    def write(self, file_obj, file_format):
        if file_format.upper() == 'FASTA':
            write_func = write_fasta
        else:
            write_func = write_fasta
        write_func(self, file_obj)

    def is_empty(self):
        return self.__len__() < 1

    def is_aligned(self):
        if self.is_empty():
            raise ValueError("The alignment is empty.\n")
        else:
            l0 = self.get_length()
            return all(len(i) == l0 for i in self.values())

    def is_all_gap(self, pos):
        """Checks to see if a column is all gap column at position x"""
        for seq in self.values():
            if seq[pos] != '-':
                return False
        return True

    def __str__(self):
        return '\n'.join([">%s\n%s" % (k, self[k])
                          for k in sorted(self.keys())])

    def divide_to_equal_chunks(self, chunks, max_chunk_size=None):
        names = self.get_sequence_names()
        ret = []
        if max_chunk_size and len(names) / chunks > max_chunk_size:
            chunks = len(names) // max_chunk_size + 1
        for i in range(0, chunks):
            subset = names[i:len(names):chunks]
            if subset:
                subset_alg = self.get_soft_sub_alignment(subset)
            else:
                subset_alg = None
            ret.append(subset_alg)
        return ret

    def get_soft_sub_alignment(self, sub_key):
        """
        Returns a read-only sub-alignment, which won't consume extra memory,
        since it will not hold a separate copy of the alignment.
        """
        return ReadonlySubalignment(sub_key, self)

class MutableAlignment(dict, ReadOnlyAlignment, object):
    """ An alignment object, that can be modified. This is the class that
    should be used mainly for holding alignments.
    """
    def __init__(self):
        """creates an empty matrix"""
        dict.__init__(self)
        self.datatype = None

    def set_alignment(self, alignment):
        for name, seq in alignment.items():
            self[name] = seq.upper()

    def read_filepath(self, filename, file_format='FASTA'):
        """Augments the matrix by reading the filepath.
        If duplicate sequence names are encountered then the old name will
        be replaced.
        """
        file_obj = open(filename, 'r')
        return self.read_file_object(file_obj, file_format=file_format)

    def read_file_object(self, file_obj, file_format='FASTA'):
        """Augments the matrix by reading the file object.
        If duplicate sequence names are encountered then the old name will
        be replaced.
        """
        if file_format.upper() == 'FASTA':
            read_func = read_fasta
#        elif (file_format.upper() == 'NEXUS'):
#            read_func = read_nexus
#        elif (file_format.upper() == 'PHYLIP'):
#            read_func = read_phylip
        else:
            raise NotImplementedError(
                "Unknown file format (%s) is not supported" % file_format)
        for name, seq in read_func(file_obj):
            self[name] = seq.upper()
        return self

    def add_column(self, pos, char='-'):
        # _LOG.debug("Added a column to reference alignment at position
        # %d" %pos)
        for name, seq in self.items():
            if hasattr(char, "get"):
                c = char.get(name, '-')
            else:
                c = char
            seq = seq[:pos] + c + seq[pos:]
            self[name] = seq

    def remove_column(self, pos):
        for name, seq in self.items():
            seq = seq[:pos] + seq[pos + 1:]
            self[name] = seq

    def remove_columns(self, indexes):
        for name, seq in self.items():
            self[name] = ''.join((
                char for idx, char in enumerate(seq) if idx not in indexes))

    def keep_columns(self, indexes):
        for name, seq in self.items():
            self[name] = ''.join((
                char for idx, char in enumerate(seq) if idx in indexes))

    def get_all_gap_cols(self):
        all_gaps = list(range(0, self.get_length()))
        for seq in self.values():
            all_gaps[:] = [i for i in all_gaps if seq[i] == '-']
        return all_gaps

    def get_all_nongap_cols(self):
        all_gaps = list(range(0, self.get_length()))
        for seq in self.values():
            all_gaps[:] = [i for i in all_gaps if seq[i] == '-']
        return [x for x in range(0, self.get_length()) if x not in all_gaps]

    def delete_all_gap(self):
        """
        Delete all sites that consists of nothing but gaps
        """
        # pdb.set_trace()

        rem = set(self.get_all_gap_cols())
        subset = [x for x in range(0, self.get_length()) if x not in rem]
        self.remove_columns(set(rem))
        #_LOG.debug("Alignment length reduced to %d" % len(subset))
        return subset

    def degap(self):
        """
        remove all
        """
        for name, seq in self.items():
            self[name] = self[name].replace("-", "")

    def get_hard_sub_alignment(self, sub_keys):
        """Creates a new alignment with a subset of the taxa."""
        new_alignment = MutableAlignment()
        new_alignment.datatype = self.datatype
        for key in sub_keys:
            new_alignment[key] = self[key]
        new_alignment.delete_all_gap()
        return new_alignment

class ReadonlySubalignment(ReadOnlyAlignment):
    """
    A class that emulates a subalignment of a given alignment. This is a
    readonly alignment and does not actually hold sequences in memory. It
    simply keeps a set of sequences that belong to the sub alignment,
    and implements methods of alignment class (and dictionaries) such that
    only subalignment keys are returned.
    """
    def __init__(self, keys, parent_alignment):
        self.seq_names = set(keys)
        self.parent_alignment = parent_alignment

    def __getitem__(self, key):
        if key in self.seq_names:
            return self.parent_alignment[key]
        else:
            raise KeyError(key)

    def __len__(self):
        return len(self.seq_names)

    def __iter__(self):
        for key in self.seq_names:
            yield key

    def get_mutable_alignment(self):
        ret = MutableAlignment()
        ret.set_alignment(self)
        return ret


class _AlignmentLookupHelper(object):
    """
    Internal helper calss
    """
    def __init__(self, pos, ref):
        self.pos = pos
        self.ref = ref

    def get(self, key, default=None):
        if key in self.ref:
            return self.ref[key][self.pos]
        else:
            return default

class ExtendedAlignment(MutableAlignment):
    """
    This is used to keep an extended alignment. An extended alignment
    has a bunch of sequences that are labeled as fragments. More importantly,
    columns of extended alignments are labeled with numbers and also it is
    known whether a column is an "insertion" column or a normal column.
    1) these can be read from .sto files.
    2) these alignments can be merged together.
    """
    def __init__(self, fragment_names):
        MutableAlignment.__init__(self)
        self.fragments = set(fragment_names)
        self._col_labels = []

    def set_alignment(self, alignment):
        MutableAlignment.set_alignment(self, alignment)
        self._reset_col_names()

    def read_file_object(self, file_obj, file_format='FASTA'):
        """ currently supports only fasta"""
        ret = MutableAlignment.read_file_object(self, file_obj, file_format)
        self._reset_col_names()
        return ret

    ############################################################
    # Added by Chengze Shen @ 2.8.2024
    #   - additional utility functions
    def is_empty(self):
        return self.__len__() < 1

    def is_aligned(self):
        if self.is_empty():
            raise ValueError("The alignment is empty.\n")
        else:
            v = list(self.values())
            first_seq_len = len(v[0])
            return all([len(i) == first_seq_len for i in v])

    def sequence_length(self):
        if self.is_aligned():
            return len(list(self.values())[0])

    def sub_alignment(self, sub_keys):
        "Creates an new alignment with a subset of the taxa."
        new_alignment = ExtendedAlignment([])
        for key in sub_keys:
            if key in self:
                new_alignment[key] = self[key]
        return new_alignment
    ############################################################

    def remove_missing_fragments(self):
        for frag in list(self.get_fragment_names()):
            if frag not in self:
                self.fragments.remove(frag)

    def add_column(self, pos, char='-', new_label=None):
        """
        A new column is added to the alignment. The new label can be value,
        or one of three directives (see below).
        """
        MutableAlignment.add_column(self, pos, char)
        if new_label == "MAX":
            self._col_labels.insert(pos, max(self._col_labels) + 1)
        elif new_label == "INC_LAST":
            self._col_labels.append(max(self._col_labels) + 1)
        elif new_label == "RESET":
            self._reset_col_names()
        else:
            self._col_labels.insert(pos, new_label)

    def remove_column(self, pos, labels="REMOVE"):
        """
        Remove a column and potentially adjust column names.
        """
        MutableAlignment.remove_column(self, pos)
        if labels == "RESET":
            self._reset_col_names()
        elif labels == "REMOVE":
            self._col_labels = self._col_labels[:pos] + \
                self._col_labels[pos + 1:]

    def _get_col_labels(self):
        return self._col_labels
    col_labels = property(_get_col_labels)

    def _reset_col_names(self):
        """ sequentially label columns"""
        self._col_labels = list(range(0, self.get_length()))

    def get_fragments_readonly_alignment(self):
        """
        Return a readonly alignment that contains only the fragments.
        """
        return ReadonlySubalignment(self.fragments, self)

    def get_base_readonly_alignment(self):
        """
        Returns a readonly subalignment that does not contain fragments.
        """
        return ReadonlySubalignment(self.get_base_seq_names(), self)

    def get_fragment_names(self):
        return self.fragments

    def get_base_seq_names(self):
        return list(set(self.keys()) - self.fragments)

    def _read_sto(self, handle):
        """
        Reads a sto file and populates this current object. Figures out
        insertion columns by finding lower case letters, asterisks, and dots.
        """
        p = re.compile(r'[a-z*.]')
        insertions = []
        last_start_ind = -1
        for line in handle:
            line = line.strip()
            if line == "//":
                # End of the alignment. ignore meta-data
                break
            elif line == "" or line[0] == "#":
                pass
            else:  # not a comment
                key, seq = line.split()
                current = self.get(key, "")
                startind = len(current)
                # finding insertion columns in limited to first sequence
                if startind != last_start_ind:
                    # s = sum(len(x) for x in current)
                    insertions.extend([m.start() + startind
                                       for m in p.finditer(seq)])
                    last_start_ind = startind
                self[key] = current + seq.replace(".", "-")
#        for k,v in self.items():
#            self[k] = "".join(v)
        self._reset_col_names()
        return set(insertions)

    '''
    6.9.2022 - added by Chengze Shen
    a new function specifically dealing with sub-alignments of WITCH
    '''
    def read_query_alignment(self, query_name, path, aformat='fasta'):
        insertions = []
        if aformat == 'fasta':
            entries = [(n, s) for n, s in read_fasta(path)]
        elif isinstance(path, MutableAlignment):
            entries = [(n, s) for n, s in path.items()]

        num_elem_per_col = [0 for _i in range(len(entries[0][1]))]

        # count how many non-gaps in each col
        query_entry = None
        for i in range(0, len(entries), 1):
            name, entry = entries[i]
            if name != query_name:
                entry_count = tuple(1 if c != '-' else 0 for c in entry)
                num_elem_per_col = list(map(add, num_elem_per_col, entry_count))
            else:
                query_entry = entry
        if not query_entry:
            print('query entry {} missing in {}???'.format(query_name, path))

        # go over the query entry and see which column it has non-gap char
        # but backbone is gap (i.e., insertion)
        # additionally, record the columns that query aligns to
        query_entry = [c for c in query_entry]
        query_aligned_columns = []
        for j in range(0, len(query_entry), 1):
            if num_elem_per_col[j] == 0 and query_entry[j] != '-':
                query_entry[j] = query_entry[j].lower()
                insertions.append(j)
                query_aligned_columns.append(-1)
            elif num_elem_per_col[j] != 0 and query_entry[j] != '-':
                query_aligned_columns.append(j)
        self.fragments.add(query_name); self[query_name] = ''.join(query_entry)
        self._reset_col_names()

        # rename all columns to sequential numbers
        insertion = -1
        for c in insertions:
            self._col_labels[c] = insertion
            insertion -= 1
        regular = 0
        for c in range(0, self.get_length()):
            if self._col_labels[c] >= 0:
                self._col_labels[c] = regular
                regular += 1
        return query_name, set(insertions), tuple(query_aligned_columns)

    def build_extended_alignment(self, base_alignment, path_to_sto_extension,
                                 convert_to_string=True):
        """
        Given a base alignment and a path to an .sto file (or a list of paths),
        this methods populates self with an extended alignment by first reading
        the base alignment, and then merging in all the .sto extension
        alignments.
        Note that the .sto alignments should contain only fragments, and also
        note that there should be no taxon overlap among extensions, or between
        extensions and the base.
        """
        if isinstance(base_alignment, Alignment):
            self.set_alignment(deepcopy(base_alignment))
        elif isinstance(base_alignment, str):
            #_LOG.debug("Reading base alignment: %s." % base_alignment)
            self.read_filepath(base_alignment, "FASTA")

        if isinstance(path_to_sto_extension, str):
            paths = [path_to_sto_extension]
        else:
            paths = path_to_sto_extension

        self.from_string_to_bytearray()
        for path in paths:
            #_LOG.debug("Reading sto extended alignment: %s." % path)
            ext = ExtendedAlignment(self.fragments)
            ext.read_extended_alignment(path)
            #_LOG.info(
            #    "Merging extension sto file (%s) into base alignment (%s)." %
            #    (path, base_alignment))
            self.merge_in(ext, False)
            #_LOG.debug(
            #    ("Finished merging extension sto file (%s) into base "
            #     "alignment (%s).") %
            #    (path, base_alignment))
            del ext
        if convert_to_string:
            self.from_bytearray_to_string()

    def read_extended_alignment(self, path, aformat="stockholm",
                                assertion=False):
        """ Reads alignment from given path and figures out "insertion"
        columns. Labels insertion columns with special labels and labels the
        rest of columns (i.e. original columns) sequentially.
        """
        handle = open(path, 'r')
        insertions = None
        if aformat.lower() == "stockholm":
            insertions = self._read_sto(handle)
            #_LOG.debug("%s insertions: %d" % (path, len(insertions)))
        else:
            raise ValueError("format %s is not supported yet" % aformat)
        assert insertions is not None

        '''Assert that insertion columns have only gaps in original seqs and
        give them appropriate labels'''
        insertion = -1
        for c in insertions:
            if assertion:
                for k in self.get_base_seq_names():
                    assert not self[k][c] != "-", \
                        ("Insertion column has sequence among original "
                         "sequences. An error? column: %d k= %s" % (c, k))
            self.col_labels[c] = insertion
            insertion -= 1
        ''' Adjust labels of other columns '''
        i = 0
        for c in range(0, self.get_length()):
            if self.col_labels[c] >= 0:
                self.col_labels[c] = i
                i += 1

    def _is_insertion_label(self, i):
        """ This differentiates between an insertion and an original column
        """
        return i < 0

    def is_insertion_column(self, col):
        return self._is_insertion_label(self.col_labels[col])

    def relabel_original_columns(self, original_labels):
        """
        This methods relabels non-insertion columns in self based on the
        input labels. Insertion column labels will not be affected.
        """
        j = 0
        #_LOG.debug(
        #    "Relabeling %d (%d) with %d labels." %
        #    (self.get_length(), len(self._col_labels), len(original_labels)))
        for i in range(0, self.get_length()):
            if not self._is_insertion_label(self.col_labels[i]):
                assert \
                    j < len(original_labels), \
                    ("Not enough labels %d %d %d.\n %s" %
                        (i, j, len(original_labels), str(self._col_labels)))
                self.col_labels[i] = original_labels[j]
                j += 1
        assert\
            j == len(original_labels), \
            ("Some of original labels are unused."
             " Some columns from original alignment went missing? %d %d" %
                (j, len(original_labels)))

    def get_insertion_columns(self):
        return [i
                for (i, x)
                in enumerate(self.col_labels)
                if self._is_insertion_label(x)]

    def get_insertion_column_ranges(self):
        pos = []
        strt = None
        prev = None
        for p in self.get_insertion_columns():
            if prev is None or prev + 1 != p:
                if strt is not None:
                    pos.append((strt, prev))
                strt = p
            prev = p
        if strt is not None:
            pos.append((strt, prev))
        return pos

    def write_insertion_column_indexes(self, path):
        file_obj = open(path, 'w')
        r = ','.join("%d-%d" % (pair[0], pair[1])
                     for pair in self.get_insertion_column_ranges())
        file_obj.write(r)
        file_obj.write("\n")
        file_obj.close()

    def remove_insertion_columns(self):
        """
        Outputs a new alignment with insertion columns masked out.
        """
        cols = self.get_insertion_columns()
        s = []
        a = 0
        for b in cols:
            if b > a:
                s.append((a, b))
            a = b + 1
        s.append((a, len(self.col_labels)))
        for name, seq in list(self.items()):
            news = []
            for c in s:
                news.append(seq[c[0]:c[1]])
            self[name] = "".join(news)

    def write_insertion_maked_to_file(self, path):
        cols = self.get_insertion_columns()
        s = []
        a = 0
        for b in cols:
            if b > a:
                s.append((a, b))
            a = b + 1
        s.append((a, len(self.col_labels)))
        file_obj = open(path, 'w')
        for name, seq in self.items():
            file_obj.write('>%s\n' % name)
            for c in s:
                file_obj.write(seq[c[0]:c[1]])
            file_obj.write("\n")
        file_obj.close()

    def from_bytearray_to_string(self):
        for k, v in self.items():
            self[k] = v.decode()

    def from_string_to_bytearray(self):
        for k, v in self.items():
            self[k] = bytearray(v, encoding="utf8")

    def merge_in(self, other, convert_to_string=True):
        """
        Merges another alignment in with the current alignment. The other
        alignment needs to be an ExtendedAlignment as well. Since both
        alignments are extended alignments, we know column labels, and we also
        know which columns are insertions. Columns with the same labels are
        merged together. In other cases gaps are introduced as necessary to
        merge the two alignments. Insertion columns are considered different
        even when they have the same label.

        convert_to_string is by default true, meaning that in the beginning
        sequences of self are assumed to be string objects. These are converted
        to bytearray for better speed, but at the end of the merge, are
        converted back to string. When convert_to_string is False, no
        conversion back and from string is performed, *but* everything is
        assumed to be in bytearrays.
        So, self is assumed to be in bytearray before calling merge, and will
        remain in bytearray after calling merge. This is useful in cases where
        multiple alignments are merged in with self. Initially, everything
        should be turned into bytearray, and after all merging is finished,
        everything can be converted back to string (using
        from_bytearray_to_string and from_string_to_bytearray).
        """
        assert isinstance(other, ExtendedAlignment)
        #_LOG.debug("Merging started ...")
        if other.is_empty():
            return
        me = 0
        she = 0  # Assumption: alignments are female!
        me_len = self.get_length() if not self.is_empty() else 0
        she_len = other.get_length()
        insertion = -1

        merged_insertion_columns = 0

        ''' Add sequences from her to my alignment '''
        for f in other.fragments:
            self.fragments.add(f)
        if convert_to_string:
            self.from_string_to_bytearray()

        selfother = {}
        for k, v in other.items():
            # assert(k not in self,
            # "Merging overlapping alignments not implemented")
            if k not in self:
                selfother[k] = bytearray(v, encoding="utf8")
        while True:
            ''' Check exit conditions'''
            if me == me_len and she == she_len:
                break

            ''' Check the 5 possible statuses between she and I '''
            if she != she_len and other.is_insertion_column(she):
                if me != me_len and self.is_insertion_column(me):
                    ''' We both have a series of insertion columns'''
                    start = me
                    while(me != me_len and self.is_insertion_column(me) and
                          she != she_len and other.is_insertion_column(she)):
                        me += 1
                        she += 1
                        merged_insertion_columns += 1
                    run = me - start
                    self.col_labels[start:me] = list(range(
                        insertion, insertion-run, -1))
                else:
                    ''' Hers is a series of insertion columns'''
                    start = she
                    while she != she_len and other.is_insertion_column(she):
                        she += 1
                    run = she - start
                    ins = bytearray(b"-") * run
                    for seq in self.values():
                        seq[me:me] = ins
                    self._col_labels[me:me] = list(range(
                        insertion, insertion - run, -1))
                    insertion -= run
                    me += run
                    me_len += run
            elif me != me_len and self.is_insertion_column(me):
                ''' Mine is a series of insertion column'''
                start = me
                while me != me_len and self.is_insertion_column(me):
                    me += 1
                run = me - start
                ins = bytearray(b"-") * run
                for v in selfother.values():
                    v[start:start] = ins
                self.col_labels[start:me] = list(
                    range(insertion, insertion-run, -1))
                insertion -= run
            elif(she == she_len or (me != me_len and
                 self.col_labels[me] < other.col_labels[she])):
                ''' My column is not present (i.e. was allgap) in the
                    "other"'''
                start = me
                while(me < me_len and (she == she_len or me != me_len and
                      self.col_labels[me] < other.col_labels[she])):
                    me += 1
                run = me - start
                ins = bytearray(b"-") * run
                for v in selfother.values():
                    v[start:start] = ins
            elif(me == me_len or (she != she_len and
                 self.col_labels[me] > other.col_labels[she])):
                ''' Her column is not present (i.e. was allgap) in "me"'''
                start = she
                while(she < she_len and (me == me_len or she != she_len and
                      self.col_labels[me] > other.col_labels[she])):
                    she += 1
                run = she - start
                ins = bytearray(b"-") * run
                for seq in self.values():
                    seq[me:me] = ins
                self._col_labels[me:me] = other.col_labels[start:she]
                me += run
                me_len += run
            elif self.col_labels[me] == other.col_labels[she]:
                ''' A shared column'''
                while(me < me_len and she < she_len and
                      self.col_labels[me] == other.col_labels[she]):
                    she += 1
                    me += 1
            else:
                raise "hmmm, we thought this should be impossible? %d %d" % (
                    me, she)

        self.update(selfother)

        if convert_to_string:
            self.from_bytearray_to_string()
        #_LOG.debug("Merging finished ...")

        return merged_insertion_columns


################## Adapted from SEPP/jobs.py ####################
def readHMMSearch(path):
    assert isinstance(path, str), "Need to give me a path to HMMSearch result"
    assert os.path.isfile(path), "File path does not exist"
    f = open(path, 'r')
    lines = f.read().split('\n')

    results = {}
    pattern = re.compile(
    r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+"
    r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)")
    start_reading = False

    for line in lines:
        line = line.strip()
        # start reading when we see "E-value"
        if not start_reading and line.startswith("E-value") is True:
            start_reading = True
        # end reading if encounter empty line
        elif start_reading and line == "":
            start_reading = False
            break
        # extend the reading
        elif start_reading:
            matches = pattern.search(line)
            if matches is not None and matches.group(0).find("--") == -1:
                results[matches.group(9).strip()] = (
                        float(matches.group(1).strip()),
                        float(matches.group(2).strip()))
    f.close()
    return results

'''
Given an aligned string (represented by upper/lower case letters and gaps,
return a condensed version that have the lowercase letters from both sides
compressed to front/back.
'''
def compressInsertions(seq):
    p = re.compile(r'[A-Z]+')
    alns = [(m.start(), m.end()) for m in p.finditer(seq)]
    # do not perform such task if there is no aligned column at all
    if len(alns) == 0:
        return seq

    # first occurrence of aligned position defines the back of front search space
    # i.e., [start, end)
    f_start, f_end = 0, alns[0][0]
    f_len = f_end - f_start

    # last occurrence of aligned position defines the front of the back search space
    b_start, b_end = alns[-1][1], len(seq)
    b_len = b_end - b_start

    # simplest way of compression: remove all gaps and add them back
    f_str_ins = seq[f_start:f_end].replace('-', '')
    f_len_ins = len(f_str_ins)
    f_str = f_str_ins + '-' * (f_len - f_len_ins)

    b_str_ins = seq[b_start:b_end].replace('-', '')
    b_len_ins = len(b_str_ins)
    b_str = '-' * (b_len - b_len_ins) + b_str_ins

    # combine the compressed front/back with the remaining sequence
    combined = f_str + seq[f_end:b_start] + b_str

    return combined



'''
A collection of jobs for TIPP3.

Jobs are designed to be run standalone, as long as all parameters
are provided correctly.
'''

import os, shutil, subprocess, stat, re, traceback, shlex
import threading
from subprocess import Popen
from abc import abstractmethod

from tipp3 import get_logger
from tipp3.configs import Configs

_LOG = get_logger(__name__)

'''
function to streamline the logging of stdout and stderr output from a job run
to a target logging file
'''
def stream_to_file(stream, fptr, logging):
    if logging:
        for line in iter(stream.readline, ''):
            fptr.write(line)
            fptr.flush()
    stream.close()

'''
Template Class Job for running external jobs
'''
class Job(object):
    def __init__(self):
        self.job_type = ""
        self.errors = []
        self.b_ignore_error = False
        self.pid = -1
        self.returncode = 0

    def __call__(self):
        return self.run()

    def get_pid(self):
        return self.pid

    # run the job with given invocation defined in a child class
    # raise errors when encountered
    def run(self, stdin="", lock=None, logging=False, shell=False):
        try:
            cmd, outpath = self.get_invocation()
            _LOG.debug(f"Running job_type: {self.job_type}, output: {outpath}")

            # failsafe for NotImplemented jobs
            if len(cmd) == 0:
                raise ValueError(
                        f"{self.job_type} does not have a valid run command. "
                        "It might be due to (invalid input types, ).")

            binpath = cmd[0]
            # special case for "java -jar ..."
            if binpath == 'java':
                binpath = cmd[2]

            assert os.path.exists(binpath) or binpath == "gzip", \
                    ("executable for %s does not exist: %s" % 
                     (self.job_type, binpath))
            assert \
                (binpath.count("/") == 0 or os.path.exists(binpath)), \
                ("path for %s does not exist (%s)" %
                 (self.job_type, binpath))

            _LOG.debug("Arguments: %s", " ".join(
                (str(x) if x is not None else "?NoneType?"
                 for x in cmd)))
        
            # logging to a local file at the target outdir
            # logname: <outdir>/{self.job_type}.txt
            stdout, stderr = '', ''
            scmd = ' '.join(cmd)
            if logging:
                logpath = os.path.join(
                        os.path.dirname(outpath), f'{self.job_type}.txt')
                outlogging = open(logpath, 'w', 1)

                # deal with piping between multiple commands (if any)
                if '|' in scmd:
                    _stdout = subprocess.PIPE
                    subcmds = [shlex.split(x) for x in scmd.split('|')]
                    prev_p = Popen(subcmds[0], text=True, bufsize=1,
                            stdout=_stdout)
                    for i in range(1, len(subcmds)):
                        if i == len(subcmds) - 1:
                            _stdout = outlogging

                        p = Popen(subcmds[i], text=True, bufsize=1,
                            stdin=prev_p.stdout,
                            stdout=_stdout, stderr=subprocess.PIPE)
                        self.pid = p.pid
                        prev_p = p
                    stdout, stderr = p.communicate()
                    stdout = ''
                else:
                    p = Popen(cmd, text=True, bufsize=1,
                            stdin=subprocess.PIPE,
                            stdout=outlogging, stderr=subprocess.PIPE)
                    self.pid = p.pid
                    stdout, stderr = p.communicate(input=stdin)
                    stdout = ''
                outlogging.close()
            else:
                if '|' in scmd:
                    subcmds = [shlex.split(x) for x in scmd.split('|')]
                    prev_p = Popen(subcmds[0], text=True, bufsize=1,
                            stdout=subprocess.PIPE)
                    for i in range(1, len(subcmds)):
                        p = Popen(subcmds[i], text=True, bufsize=1,
                            stdin=prev_p.stdout,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        self.pid = p.pid
                        prev_p = p
                    stdout, stderr = p.communicate()
                else:
                    p = Popen(cmd, text=True, bufsize=1,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    self.pid = p.pid
                    stdout, stderr = p.communicate(input=stdin)

            self.returncode = p.returncode

            #p = Popen(cmd, bufsize=1, text=True,
            #        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            #_outdir = os.path.dirname(os.path.realpath(outpath))
            #if logging:
            #    logfile = os.path.join(_outdir, 'runtime.txt')
            #    fptr = open(logfile, 'w', buffering=1)

            #    # initialize writing to logging file (if logging)
            #    stdout_thread = threading.Thread(target=stream_to_file,
            #            args=(p.stdout, fptr, logging))
            #    stderr_thread = threading.Thread(target=stream_to_file,
            #            args=(p.stderr, fptr, logging))
            #    stdout_thread.start()
            #    stderr_thread.start()

            #    # join threads
            #    stdout_thread.join()
            #    stderr_thread.join()

            # finish up the process run
            #stdout, stderr = p.communicate()
            #self.returncode = p.returncode

            if self.returncode == 0:
                if lock:
                    try:
                        lock.acquire()
                        _LOG.debug(f"{self.job_type} completed, output: {outpath}")
                    finally:
                        lock.release()
                else:
                    _LOG.debug(f"{self.job_type} completed, output: {outpath}")
                return outpath
            else:
                error_msg = ' '.join([f"Error occurred running {self.job_type}.",
                    f"Return code: {self.returncode}"])
                print(error_msg)
                if lock:
                    try:
                        lock.acquire()
                        _LOG.error(error_msg + '\nSTDOUT: ' + stdout +
                                '\nSTDERR: ' + stderr)
                    finally:
                        lock.release()
                else:
                    _LOG.error(error_msg + '\nSTDOUT: ' + stdout +
                            '\nSTDERR: ' + stderr)
                #_LOG.error(error_msg + '\n' + stdout)
                exit(1)
        except Exception:
            traceback.print_exc()
            raise

    # implement in child class
    # return: (cmd, outpath)
    @abstractmethod
    def get_invocation(self):
        raise NotImplementedError(
            "get_invocation() should be implemented by subclasses")

'''
A BLASTN job that will run BLASTN to bin reads against the reference package
marker genes
'''
class BlastnJob(Job):
    def __init__(self, **kwargs):
        Job.__init__(self)
        self.job_type = 'blastn'
        self.outfmt = 0

        # arguments for running BLASTN
        self.path = '' 
        self.query_path = ''
        self.database_path = ''
        self.outdir = ''
        self.num_threads = 1 

        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_invocation(self):
        #blast:database = blast/alignment.fasta.db
        #blast:seq-to-marker-map = blast/seq2marker.tab
        self.outpath = os.path.join(self.outdir, 'blast.alignment.out') 

        # check input type: if .fasta as suffix then it is fine
        # if .fasta.gz suffix, then use the file as stdin 
        name_parts = self.query_path.split('.')
        suffix = name_parts[-1]

        cmd = []

        # normal input type with fasta/fa files
        if suffix in ['fa', 'fasta']:
            cmd = [self.path, '-db', self.database_path,
                    '-outfmt', str(self.outfmt),
                    '-query', self.query_path,
                    '-out', self.outpath,
                    '-num_threads', str(self.num_threads)]
        # awk to process fq and fastq files as input
        elif suffix in ['fq', 'fastq']:
            cmd.extend(['awk',
                '\'NR%4==1 {print \">\"substr($0, 2)} NR%4==2 {print $0}\'',
                '|'])
        # gzip to process gzip and gz files as input
        elif suffix in ['gz', 'gzip']: 
            cmd.extend(['gzip', '-dc', self.query_path, '|'])
            # check if this is fasta/fa gzip or fastq/fq gzip
            if len(name_parts) > 2:
                suffix2 = name_parts[-2]

            # if we have another layer of fastq/fq to deal with, extend
            # the cmd to deal with that
            if suffix2 in ['fastq', 'fq']:
                cmd.extend(['awk',
                    '\'NR%4==1 {print \">\"substr($0, 2)} NR%4==2 {print $0}\'',
                    '|'])

            # piped query will go in to BLAST as stdin
            cmd.extend([self.path, '-db', self.database_path,
                    '-outfmt', str(self.outfmt),
                    '-query', '-',
                    '-out', self.outpath,
                    '-num_threads', str(self.num_threads)])
        # will raise ValueError when run, due to not having a recognizable
        # input type
        else:
            return [], self.outpath
        
        return cmd, self.outpath

'''
A WITCH alignment job that will run WITCH to align a set of query reads
to their target marker gene backbone alignment
'''
class WITCHAlignmentJob(Job):
    def __init__(self, **kwargs):
        Job.__init__(self)
        self.job_type = 'witch-alignment'

        # initialize parameters
        self.path = ''
        self.query_path = ''
        self.backbone_path = ''
        self.backbone_tree_path = ''
        self.outdir = ''
        self.num_cpus = 1

        # set parameters
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.kwargs = kwargs
    
    def get_invocation(self):
        self.outpath = os.path.join(self.outdir, 'est.aln.masked.fasta')
        cmd = [self.path, '-o', 'est.aln.fasta',
                ]
        # extend from additional kwargs
        for k, v in self.kwargs.items():
            if k != 'path':
                param_name = k.replace('_', '-')
                cmd.extend([f"--{param_name}", str(v)])
        return cmd, self.outpath

'''
A BSCAMPP job that will run BSCAMPP for placing aligned query reads 
'''
class BscamppJob(Job):
    def __init__(self,
            path, query_alignment_path, backbone_alignment_path,
            backbone_tree_path, tree_model_path, outdir, num_cpus,
            **kwargs):
        Job.__init__(self)
        self.job_type = 'bscampp'
        
        # initialize parameters
        self.path = path
        self.query_alignment_path = query_alignment_path
        self.backbone_alignment_path = backbone_alignment_path
        self.backbone_tree_path = backbone_tree_path
        self.tree_model_path = tree_model_path
        self.outdir = outdir
        self.num_cpus = num_cpus
        self.kwargs = kwargs

    def get_invocation(self):
        self.outpath = os.path.join(self.outdir, 'placement.jplace')
        cmd = [self.path,
                '-q', self.query_alignment_path,
                '-a', self.backbone_alignment_path,
                '-t', self.backbone_tree_path,
                '-i', self.tree_model_path,
                '-d', self.outdir,
                '-o', 'placement',
                '--num-cpus', str(self.num_cpus),
                ]
        # extend any additional kwargs specified for BSCAMPP
        for k, v in self.kwargs.items():
            # special case: not processing support_value (used later)
            if k == 'support_value':
                continue
            param = k.replace('_', '-')
            cmd.extend([f'--{param}', str(v)])
        return cmd, self.outpath 

'''
A pplacer-taxtastic job that runs pplacer with the taxtastic refpkg
'''
class PplacerTaxtasticJob(Job):
    def __init__(self,
            path, query_alignment_path, refpkg_path, outdir, num_cpus,
            **kwargs):
        Job.__init__(self)
        self.job_type = 'pplacer-taxtastic'

        self.path = path
        self.query_alignment_path = query_alignment_path
        self.refpkg_path = refpkg_path
        self.outdir = outdir
        self.num_cpus = num_cpus
        self.model_type = 'GTR'
        self.kwargs = kwargs

    def get_invocation(self):
        self.outpath = os.path.join(self.outdir, 'placement.jplace')
        cmd = [self.path,
                '-m', self.model_type,
                '-c', self.refpkg_path,
                '-o', self.outpath,
                '-j', str(self.num_cpus),
                self.query_alignment_path,
                ]
        ## extend any additional kwargs specified for BSCAMPP
        #for k, v in self.kwargs.items():
        #    # special case: not processing support_value (used later)
        #    if k == 'support_value':
        #        continue
        #    param = k.replace('_', '-')
        #    cmd.extend([f'--{param}', str(v)])
        return cmd, self.outpath

'''
A JsonMerger job for TIPP to read taxonomic information from a .jplace file
with the corresponding taxonomic tree
'''
class TIPPJsonMergerJob(Job):
    def __init__(self, **kwargs):
        Job.__init__(self)
        self.job_type = 'tippjsonmerger'

        self.path = ''
        self.taxonomy_path = ''
        self.mapping_path = ''
        self.classification_path = ''
        self.outdir = ''

        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_invocation(self):
        self.outpath = os.path.join(self.outdir, 'tippjsonmerger.jplace') 
        cmd = ['java', '-jar', self.path,
                '-', '-', self.outpath,
                '-t', self.taxonomy_path, '-m', self.mapping_path,
                '-p', '0.0', '-C', '0.0', '-c', self.classification_path]
        return cmd, self.classification_path

'''
A generic job that is designed for univeral types of additional jobs that
a user/coder can add to TIPP3. just need to supplement the correct configuration
in main.config or user.config
NOTE: the binary executable path given needs to be executable
'''
class GenericJob(Job):
    def __init__(self, **kwargs):
        Job.__init__(self)
        self.job_type = 'generic'

        # note: [job_type, path, outpath] needs to be set from configuration file
        for k, v in kwargs.items():
            setattr(self, k, v)
    
    def get_invocation(self):
        outpath, cmd = None, None 
        if getattr(self, 'outpath') != None:
            outpath = self.outpath
        if getattr(self, 'cmd') != None:
            cmd = cmd.split()
        
        # raise Error for generic job that does not have a command 
        if not outpath or not cmd:
            _LOG.error(f"Generic Job {self.job_type} " 
                "do not have outpath and cmd in configuration.")
            exit(1)

        return cmd, outpath



from witch_msa.helpers.alignment_tools import Alignment
from pyhmmer import easel

# helper function to convert an alignment object to TextMSA object
def alignmentToTextMSA(aln, name):
    sequences = []
    for taxon, seq in aln.items():
        sequences.append(easel.TextSequence(name=taxon.encode(), sequence=seq))
    msa = easel.TextMSA(name=name.encode(), sequences=sequences)
    return msa

# helper function to obtain alphabet given molecule type
def moleculeToAlphabet(molecule):
    alphabet = None
    if molecule == 'amino':
        alphabet = easel.Alphabet.amino()
    elif molecule == 'dna':
        alphabet = easel.Alphabet.dna()
    elif molecule == 'rna':
        alphabet = easel.Alphabet.rna()
    else:
        raise ValueError(f"alphabet {molecule} is not amino, dna, or rna")
    return alphabet



###########################################################################
#    Copyright 2012 Siavash Mirarab, Nam Nguyen, and Tandy Warnow.
#    This file is part of SEPP.
#
#    SEPP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SEPP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SEPP.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################

from operator import itemgetter
import logging
import os

'''
Updated @ 3.12.2025 by Chengze Shen

Major changes to suit TIPP3 pipeline.
'''
__version__ = "0.3a"
_INSTALL_PATH = __path__[0]

__all__ = ['read_binning', 'read_alignment', 'read_placement',
        'tipp3_pipeline', 'jobs', 'refpkg_downloader']

def get_setup_path():
    return _INSTALL_PATH


def get_logging_level(logging_level='info'):
    #### logging level map ####
    logging_level_map = {
            'DEBUG': logging.DEBUG, 'INFO': logging.INFO,
            'WARNING': logging.WARNING, 'ERROR': logging.ERROR,
            'CRITICAL': logging.CRITICAL,
            }
    # obtain environment variable to determine the logging level,
    # if TIPP3_LOGGING_LEVEL is not empty
    env_level = os.getenv('TIPP_LOGGING_LEVEL')
    if env_level is not None:
        ll = env_level.upper()
    else:
        ll = logging_level.upper()
    return logging_level_map.get(ll, logging.INFO) 
    #return logging.DEBUG if _DEBUG else logging.INFO

__set_loggers = set()

def get_logger(name="tipp3", log_path=None, logging_level='info'):
    logger = logging.getLogger(name)
    if name not in __set_loggers:
        level = get_logging_level(logging_level)
        logging_formatter = logging.Formatter(
            ("[%(asctime)s] %(filename)s (line %(lineno)d):"
             " %(levelname) 8s: %(message)s"))
        logging_formatter.datefmt = '%H:%M:%S'
        logger.setLevel(level)
        if log_path == None:
            ch = logging.StreamHandler()
        else:
            # use FileHandler instead
            ch = logging.FileHandler(log_path, mode='a')

        ch.setLevel(level)
        ch.setFormatter(logging_formatter)
        logger.addHandler(ch)
        __set_loggers.add(name)
    return logger


#### unused for TIPP3 for now
#def reset_loggers():
#    global __set_loggers
#    __set_loggers = set()
#    import pkgutil
#    import sepp
#    package = sepp
#    for modl, name, _ in pkgutil.iter_modules(package.__path__):
#        logger = (getattr(getattr(sepp, name, None), "_LOG", None))
#        print("--- *", name, logger)
#        if logger:
#            setattr(getattr(sepp, name, None), "_LOG", get_logger(
#                "tipp3.%s" % name))


def log_exception(logger):
    """Logs the exception trace to the logObj as an error"""
    import traceback
    import io
    s = io.StringIO()
    traceback.print_exc(None, s)
    logger.error(s.getvalue())
    exit(1)

os.sys.setrecursionlimit(1000000)


def sort_by_value(d, reverse=False):
    return sorted(iter(d.items()), key=itemgetter(1), reverse=reverse)
